/*
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * TriState.java
 * Copyright (C) 2014-2019 University of Waikato, Hamilton, New Zealand
 */
package adams.core;

import adams.core.option.AbstractOption;

/**
 * Enum that enhances a boolean to have a third state: not set.
 * 
 * @author  fracpete (fracpete at waikato dot ac dot nz)
 */
public enum TriState
  implements EnumWithCustomDisplay<TriState>{

  /** true. */
  TRUE("True"),
  /** false. */
  FALSE("False"),
  /** not set. */
  NOT_SET("Not set");

  /** the display string. */
  private String m_Display;

  /** the commandline string. */
  private String m_Raw;

  /**
   * The constructor.
   *
   * @param display	the string to use as display
   */
  private TriState(String display) {
    m_Display = display;
    m_Raw     = super.toString();
  }

  /**
   * Returns the display string.
   *
   * @return		the display string
   */
  public String toDisplay() {
    return m_Display;
  }

  /**
   * Returns the raw enum string.
   *
   * @return		the raw enum string
   */
  public String toRaw() {
    return m_Raw;
  }

  /**
   * Returns the display string.
   *
   * @return		the display string
   */
  @Override
  public String toString() {
    return toDisplay();
  }

  /**
   * Checks whether the boolean state corresponds to {@link #TRUE} or {@link #FALSE}.
   *
   * @param state 	the state to compare with
   * @return		true if "TRUE && true" or "FALSE && false" or "NOT_SET && null"
   */
  public boolean matches(Boolean state) {
    return ((this == TriState.NOT_SET) && (state == null))
      || ((this == TriState.TRUE) && state)
      || ((this == TriState.FALSE) && !state);
  }

  /**
   * Parses the given string and returns the associated enum.
   *
   * @param s		the string to parse
   * @return		the enum or null if not found
   */
  public TriState parse(String s) {
    return valueOf((AbstractOption) null, s);
  }

  /**
   * Returns the enum as string.
   *
   * @param option	the current option
   * @param object	the enum object to convert
   * @return		the generated string
   */
  public static String toString(AbstractOption option, Object object) {
    return ((TriState) object).toRaw();
  }

  /**
   * Returns an enum generated from the string.
   *
   * @param option	the current option
   * @param str	the string to convert to an enum
   * @return		the generated enum or null in case of error
   */
  public static TriState valueOf(AbstractOption option, String str) {
    TriState	result;

    result = null;

    // default parsing
    try {
      result = valueOf(str);
    }
    catch (Exception e) {
      // ignored
    }

    // try display
    if (result == null) {
      for (TriState dt: values()) {
	if (dt.toDisplay().equalsIgnoreCase(str)) {
	  result = dt;
	  break;
	}
      }
    }

    return result;
  }

}
