/*
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * AbstractBaseDateTypeEditor.java
 * Copyright (C) 2016-2019 University of Waikato, Hamilton, New Zealand
 *
 */

package adams.gui.goe;

import adams.core.DateValueSupporter;
import adams.core.base.BaseObject;
import adams.gui.chooser.DateProvider;
import adams.gui.core.BaseButton;
import adams.gui.core.BaseButtonWithDropDownMenu;
import adams.gui.core.BaseComboBox;
import adams.gui.core.BaseTextField;
import adams.gui.core.GUIHelper;
import adams.gui.help.HelpFrame;
import adams.parser.GrammarSupplier;

import javax.swing.BorderFactory;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.FlowLayout;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.lang.reflect.Array;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Map;

/**
 * Ancestor property editors that handle base date types.
 *
 * @author FracPete (fracpete at waikato dot ac dot nz)
 * @param <B> the base object type in use
 * @param <P> the panel in use for selecting the date type
 */
public abstract class AbstractBaseDateTypeEditor<B extends BaseObject & DateValueSupporter & GrammarSupplier,P extends DateProvider>
  extends AbstractPropertyEditorSupport
  implements CustomStringRepresentationHandler, MultiSelectionEditor, InlineEditorSupport {

  /**
   * The supported placeholders.
   *
   * @author FracPete (fracpete at waikato dot ac dot nz)
   * @version $Revision$
   */
  public enum Placeholder {
    INF_PAST,
    NOW,
    INF_FUTURE
  }

  /** The calendar component for editing the date. */
  protected P m_Calendar;

  /** For specific date placeholders. */
  protected BaseComboBox<String> m_ComboBoxPlaceholders;

  /** For entering a custom date string. */
  protected BaseTextField m_TextCustom;

  /** the button for the bringing up the help for the custom format. */
  protected BaseButton m_ButtonCustomHelp;

  /** The NOW button. */
  protected BaseButton m_ButtonNow;

  /** the template button. */
  protected BaseButtonWithDropDownMenu m_ButtonTemplates;

  /** The OK button. */
  protected BaseButton m_ButtonOK;

  /** for parsing the date type. */
  protected B m_Date;

  /**
   * Initializes the members.
   */
  protected void initialize() {
    super.initialize();

    m_Date = newDateType();
  }

  /**
   * Returns a new instance of the date type used for parsing.
   *
   * @return		the instance
   */
  protected B newDateType() {
    return newDateType((String) null);
  }

  /**
   * Returns a new instance of the date type.
   *
   * @param s		the string to instantiate with, can be null
   * @return		the instance
   */
  protected abstract B newDateType(String s);

  /**
   * Returns a new instance of the date type.
   *
   * @param d		the date to initialize with
   * @return		the instance
   */
  protected abstract B newDateType(Date d);

  /**
   * Returns a custom string representation of the object.
   *
   * @param obj		the object to turn into a string
   * @return		the string representation
   */
  public String toCustomStringRepresentation(Object obj) {
    return ((B) obj).getValue();
  }

  /**
   * Returns an object created from the custom string representation.
   *
   * @param str		the string to turn into an object
   * @return		the object
   */
  public Object fromCustomStringRepresentation(String str) {
    return newDateType(str);
  }

  /**
   * Returns a representation of the current property value as java source.
   *
   * @return 		a value of type 'String'
   */
  public String getJavaInitializationString() {
    String	result;

    result = "new " + getValue().getClass().getName() + "(\"" + toCustomStringRepresentation(getValue()) + "\")";

    return result;
  }

  /**
   * Paints a representation of the current Object.
   *
   * @param gfx 	the graphics context to use
   * @param box 	the area we are allowed to paint into
   */
  public void paintValue(Graphics gfx, Rectangle box) {
    FontMetrics 	fm;
    int 		vpad;
    String 		val;

    fm   = gfx.getFontMetrics();
    vpad = (box.height - fm.getHeight()) / 2;
    if (getValue() == null)
      val = AbstractPropertyEditorSupport.NULL;
    else
      val = toCustomStringRepresentation(getValue());
    if (val.isEmpty())
      val = AbstractPropertyEditorSupport.EMPTY;
    gfx.drawString(val, 2, fm.getHeight() + vpad);
  }

  /**
   * Instantiates a new panel for picking the date type.
   *
   * @return		the panel
   */
  protected abstract P newPanel();

  /**
   * Returns the string equivalent of the placeholder.
   *
   * @param ph		the placeholder to get the string representation for
   * @return		the string representation
   */
  protected abstract String getPlaceholder(Placeholder ph);

  /**
   * Returns the text to use for the "now" button.
   *
   * @return		the button text
   */
  protected abstract String getNowButtonText();

  /**
   * Gets the custom editor component.
   *
   * @return 		the editor
   */
  protected JComponent createCustomEditor() {
    JPanel	panelAll;
    JPanel	panel;
    JPanel	panelValues;
    JPanel	panelCustom;
    JPanel 	panelButtons;
    BaseButton buttonCancel;

    panelAll = new JPanel(new BorderLayout());
    panelAll.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
    panelValues = new JPanel(new BorderLayout());
    panelAll.add(panelValues, BorderLayout.CENTER);
    panelCustom = new JPanel(new BorderLayout());
    panelValues.add(panelCustom, BorderLayout.SOUTH);

    m_Calendar  = newPanel();
    panelValues.add((JPanel) m_Calendar, BorderLayout.CENTER);

    m_ComboBoxPlaceholders = new BaseComboBox<>(new String[]{
      "Custom",
      "Selected",
      getPlaceholder(Placeholder.INF_PAST),
      getPlaceholder(Placeholder.NOW),
      getPlaceholder(Placeholder.INF_FUTURE)
    });
    m_ComboBoxPlaceholders.addActionListener((ActionEvent e) -> {
      int index = m_ComboBoxPlaceholders.getSelectedIndex();
      if (index > 1) {
	B date = newDateType((String) m_ComboBoxPlaceholders.getSelectedItem());
	m_Calendar.setDate(date.dateValue());
      }
      m_TextCustom.setEnabled(index == 0);
      m_ButtonNow.setEnabled(index == 1);
      ((JPanel) m_Calendar).setEnabled(index == 1);
    });
    m_ButtonNow = new BaseButton(getNowButtonText());
    m_ButtonNow.addActionListener((ActionEvent e) -> m_Calendar.setDate(newDateType(getPlaceholder(Placeholder.NOW)).dateValue()));
    panel = new JPanel(new FlowLayout(FlowLayout.LEFT));
    panel.setBorder(BorderFactory.createEmptyBorder(5, 0, 5, 0));
    panel.add(new JLabel("Type"));
    panel.add(m_ComboBoxPlaceholders);
    panel.add(m_ButtonNow);
    panelCustom.add(panel, BorderLayout.NORTH);

    m_TextCustom = new BaseTextField(12);
    m_TextCustom.getDocument().addDocumentListener(new DocumentListener() {
      public void removeUpdate(DocumentEvent e) {
	update();
      }
      public void insertUpdate(DocumentEvent e) {
	update();
      }
      public void changedUpdate(DocumentEvent e) {
	update();
      }
      protected void update() {
	updateButtons();
      }
    });
    m_ButtonCustomHelp = new BaseButton(GUIHelper.getIcon("help.gif"));
    m_ButtonCustomHelp.addActionListener(e1 -> showHelp());
    panel = new JPanel(new FlowLayout(FlowLayout.LEFT));
    panel.setBorder(BorderFactory.createEmptyBorder(5, 0, 5, 0));
    panel.add(new JLabel("Custom date"));
    panel.add(m_TextCustom);
    panel.add(m_ButtonCustomHelp);
    panelCustom.add(panel, BorderLayout.CENTER);

    panelButtons = new JPanel(new FlowLayout(FlowLayout.RIGHT));
    panelAll.add(panelButtons, BorderLayout.SOUTH);

    m_ButtonTemplates = null;
    if (BaseObject.getTemplates(newDateType().getClass()).size() > 0) {
      m_ButtonTemplates = new BaseButtonWithDropDownMenu(GUIHelper.getIcon("template.gif"));
      m_ButtonTemplates.setToolTipText("Templates");
      Map<String,B> templates = (Map<String,B>) BaseObject.getTemplates(newDateType().getClass());
      List<String> items = new ArrayList<>(templates.keySet());
      Collections.sort(items);
      for (String item: items) {
	JMenuItem menuitem = new JMenuItem(item);
	menuitem.addActionListener((ActionEvent ae) -> setValue(templates.get(item)));
	m_ButtonTemplates.addToMenu(menuitem);
      }
      panelButtons.add(m_ButtonTemplates);
    }

    m_ButtonOK = new BaseButton("OK");
    m_ButtonOK.setMnemonic('O');
    m_ButtonOK.addActionListener((ActionEvent e) -> {
      int index = m_ComboBoxPlaceholders.getSelectedIndex();
      if (index == 0)
	setValue(newDateType(m_TextCustom.getText()));
      else if (index == 1)
	setValue(newDateType(m_Calendar.getDate()));
      else
	setValue(newDateType((String) m_ComboBoxPlaceholders.getSelectedItem()));
      closeDialog(APPROVE_OPTION);
    });
    panelButtons.add(m_ButtonOK);

    buttonCancel = new BaseButton("Cancel");
    buttonCancel.setMnemonic('C');
    buttonCancel.addActionListener((ActionEvent e) -> closeDialog(CANCEL_OPTION));
    panelButtons.add(buttonCancel);

    return panelAll;
  }

  /**
   * Sets the enabled state of the buttons.
   */
  protected void updateButtons() {
    int		index;

    index = m_ComboBoxPlaceholders.getSelectedIndex();

    if (index == 0)
      m_ButtonOK.setEnabled(m_Date.isValid(m_TextCustom.getText()));
    else
      m_ButtonOK.setEnabled(true);
  }

  /**
   * Displays the help for the
   */
  protected void showHelp() {
    HelpFrame.showHelp(m_Date.getClass(), m_Date.getGrammar(), false);
  }

  /**
   * Initializes the display of the value.
   */
  protected void initForDisplay() {
    super.initForDisplay();
    m_TextCustom.setText(((BaseObject) getValue()).getValue());
    m_TextCustom.setToolTipText(((BaseObject) getValue()).getTipText());
    m_Calendar.setDate(((B) getValue()).dateValue());
    m_ComboBoxPlaceholders.setSelectedIndex(m_ComboBoxPlaceholders.getSelectedIndex());
    updateButtons();
  }

  /**
   * Returns the selected objects.
   *
   * @param parent	the parent container
   * @return		the objects
   */
  @Override
  public Object[] getSelectedObjects(Container parent) {
    Object			result;
    MultiLineValueDialog	dialog;
    List<String>		lines;
    int				i;

    initForDisplay();

    if (GUIHelper.getParentDialog(parent) != null)
      dialog = new MultiLineValueDialog(GUIHelper.getParentDialog(parent));
    else
      dialog = new MultiLineValueDialog(GUIHelper.getParentFrame(parent));
    dialog.setInfoText("Enter the date values, one per line:");
    dialog.setLocationRelativeTo(parent);
    dialog.setVisible(true);

    lines  = dialog.getValues();
    result = Array.newInstance(m_Date.getClass(), lines.size());
    for (i = 0; i < lines.size(); i++)
      Array.set(result, i, newDateType(lines.get(i)));

    return (B[]) result;
  }

  /**
   * Checks whether inline editing is available.
   *
   * @return		true if editing available
   */
  public boolean isInlineEditingAvailable() {
    return true;
  }

  /**
   * Sets the value to use.
   *
   * @param value	the value to use
   */
  public void setInlineValue(String value) {
    if (isInlineValueValid(value))
      setValue(newDateType(value));
  }

  /**
   * Returns the current value.
   *
   * @return		the current value
   */
  public String getInlineValue() {
    return ((BaseObject) getValue()).getValue();
  }

  /**
   * Checks whether the value id valid.
   *
   * @param value	the value to check
   * @return		true if valid
   */
  public boolean isInlineValueValid(String value) {
    return ((BaseObject) getValue()).isValid(value);
  }
}
